from .preloaded import (  # noqa: F401
    load_voc2012,
    load_oxfordiiitpets,
    load_icdar2015,
    load_shapes
)
from .coco import load_coco  # noqa: F401
from .voc import load_voc  # noqa: F401
from .via import load_via  # noqa: F401